(module botan
  (botan-api-version botan-version-string botan-version
   data? data-eqv? data-scrub! hex-encode hex-decode base64-encode base64-decode
   rng? make-rng rng-get rng-reseed! rng-add-entropy!
   hash? make-hash hash-name hash-copy hash-clear! hash-update! hash-length hash-final
   mac? make-mac mac-name mac-clear! mac-keyspec mac-set-key! mac-update! mac-length mac-final
   cipher? make-cipher cipher-name cipher-clear! cipher-keyspec cipher-set-key!
   cipher-output-length cipher-tag-length
   cipher-valid-nonce-length? cipher-default-nonce-length
   cipher-transform
   pbkdf-iterations pbkdf kdf
   bcrypt-generate bcrypt-valid?
   privkey? generate-privkey
   make-privkey/rsa make-privkey/dsa make-privkey/elgamal make-privkey/dh
   make-privkey/ed25519 make-privkey/x25519 make-privkey/ecdsa make-privkey/ecdh make-privkey/sm2
   load-privkey data->privkey privkey->string privkey->blob privkey->ed25519 privkey->x25519
   privkey-check privkey-algorithm privkey-field
   pubkey? privkey->pubkey
   make-pubkey/rsa make-pubkey/dsa make-pubkey/elgamal make-pubkey/dh
   make-pubkey/ed25519 make-pubkey/x25519 make-pubkey/ecdsa make-pubkey/ecdh make-pubkey/sm2
   load-pubkey data->pubkey pubkey->string pubkey->blob pubkey->ed25519 pubkey->x25519
   pubkey-algorithm pubkey-field
   pubkey-estimated-strength pubkey-fingerprint
   pk-encrypt? make-pk-encrypt pk-encrypt-output-length pk-encrypt
   pk-decrypt? make-pk-decrypt pk-decrypt-output-length pk-decrypt
   pk-sign? make-pk-sign pk-sign-update! pk-sign-length pk-sign-final
   pk-verify? make-pk-verify pk-verify-update! pk-verify-final
   pk-key-agreement? make-pk-key-agreement pk-key-agreement-length pk-key-agreement-export-public pk-key-agreement
   x509-cert? load-x509-cert load-x509-cert-chain data->x509-cert x509-cert->string x509-cert->pubkey
   x509-cert-not-before x509-cert-not-after x509-cert-serial-number x509-cert-fingerprint
   x509-cert-issuer-key-id x509-cert-issuer-dn x509-cert-subject-key-id x509-cert-subject-dn
   x509-crl? load-x509-crl data->x509-crl
   x509-usage-allowed? x509-hostname-match? x509-revoked? x509-verify
   hotp? make-hotp hotp-generate hotp-valid?
   totp? make-totp totp-generate totp-valid?
   fpe? make-fpe fpe-encrypt fpe-decrypt)

(import
  scheme
  srfi-4
  (only srfi-1 reverse! fold)
  (only srfi-13 string-index string-every string-contains)
  (only srfi-14 char-set:letter char-set:digit char-set-union char-set-adjoin)
  (chicken blob)
  (chicken base)
  (chicken foreign)
  (chicken condition)
  (chicken memory)
  (only (chicken gc) set-finalizer!)
  (only (chicken io) read-string)
  (only (chicken time) current-seconds))

(foreign-declare
  "#include <stddef.h>"
  "#include <botan/ffi.h>")

(define ((botan-error/message location specific-message . arguments) status)
  (receive (status generic-message)
    (case status
      [(1)   (values 'invalid-verifier "invalid verifier")]
      [(-1)  (values 'invalid-input "invalid input")]
      [(-2)  (values 'bad-mac "bad message authentication code")]
      [(-10) (values 'insufficient-buffer-space "insufficient buffer space")]
      [(-20) (values 'exception-thrown "unknown C++ exception")]
      [(-21) (values 'out-of-memory "out of memory")]
      [(-30) (values 'bad-flag "bad flag")]
      [(-31) (values 'null-pointer "null pointer")]
      [(-32) (values 'bad-parameter "bad parameter")]
      [(-33) (values 'key-not-set "key not set")]
      [(-34) (values 'invalid-key-length "invalid key length")]
      [(-40) (values 'not-implemented "not implemented")]
      [(-50) (values 'invalid-object "invalid object")]
      [else  (values 'unknown-error "unknown error")])
    (abort
     (make-composite-condition
      (make-property-condition
       'exn
       'location location
       'message (or specific-message generic-message)
       'arguments arguments)
      (make-property-condition
       'botan
       'status status)))))

(define (botan-error location . arguments)
  (apply botan-error/message location #f arguments))

(define (type-error location x)
  ((botan-error location x) -50))

(define char-set:nice-symbol
  (char-set-adjoin
   (char-set-union char-set:letter char-set:digit)
   #\+ #\- #\< #\> #\/ #\: #\= #\?))

(define-syntax define-pointer-type
  (syntax-rules ()
    [(define-pointer-type (scheme-name struct-name) tag predicate)
     (begin
       (define tag
         (symbol->string 'scheme-name))
       (define (predicate x)
         (tagged-pointer? x tag))
       (define-foreign-type scheme-name (c-pointer (struct struct-name))
         (lambda (p)
           (if (or (not p) (predicate p)) p (type-error 'scheme-name p)))
         (lambda (p)
           (and p (tag-pointer p tag)))))]))

(define botan-api-version
  (case-lambda
    [()
     ((foreign-lambda unsigned-int32 "botan_ffi_api_version"))]
    [(version)
     (case ((foreign-lambda int "botan_ffi_supports_api" unsigned-int32) version)
       [(0)  (void)]
       [else => (botan-error 'botan-api-version version)])]))

(define botan-version-string
  (foreign-lambda c-string "botan_version_string"))

(define (botan-version)
  (let ([major ((foreign-lambda unsigned-int32 "botan_version_major"))]
        [minor ((foreign-lambda unsigned-int32 "botan_version_minor"))]
        [patch ((foreign-lambda unsigned-int32 "botan_version_patch"))]
        [stamp ((foreign-lambda unsigned-int32 "botan_version_datestamp"))])
    `(,major ,minor ,patch ,@(if (zero? stamp) (list) (list stamp)))))

(define (data? x)
  (or (not x) (string? x) (blob? x) (u8vector? x)))

(define (data-object/length location x)
  (cond
    [(not x)
     (values #f 0)]
    [(string? x)
     (values x (string-length x))]
    [(blob? x)
     (values x (blob-size x))]
    [(u8vector? x)
     (values (u8vector->blob/shared x) (u8vector-length x))]
    [else
     (type-error location x)]))

(define (data-eqv? a b)
  (let*-values ([(a* a-length) (data-object/length 'data-eqv? a)]
                [(b* b-length) (data-object/length 'data-eqv? b)])
    (and
      (= a-length b-length)
      (or
        (zero? a-length)
        (case ((foreign-lambda int "botan_constant_time_compare" (nonnull-scheme-pointer (const unsigned-byte)) (nonnull-scheme-pointer (const unsigned-byte)) size_t)
               a* b* a-length)
          [(0)  #t]
          [(-1) #f]
          [else => (botan-error 'data-eqv? a b)])))))

(define (data-scrub! data)
  (receive (data* data-length) (data-object/length 'data-scrub! data)
    (unless (zero? data-length)
      (case ((foreign-lambda int "botan_scrub_mem" (nonnull-scheme-pointer unsigned-byte) size_t) data data-length)
        [(0)  (void)]
        [else => (botan-error 'data-scrub! data)]))))

(define (hex-encode v #!key downcase)
  (let*-values ([(v* n) (data-object/length 'hex-encode v)]
                [(s) (make-string (* 2 n))])
    (case ((foreign-lambda int "botan_hex_encode" (scheme-pointer (const unsigned-byte)) size_t (nonnull-scheme-pointer char) unsigned-int32)
           v* n s (if downcase 1 0))
      [(0)  s]
      [else => (botan-error 'hex-encode v downcase: downcase)])))

(define (hex-decode s)
  (let* ([n (string-length s)]
         [v (make-u8vector (ceiling (/ n 2)))])
    (let-location ([o size_t (u8vector-length v)])
      (case ((foreign-lambda int "botan_hex_decode" (nonnull-scheme-pointer (const char)) size_t nonnull-u8vector (c-pointer size_t))
             s n v (location o))
        [(0)  (if (= o (u8vector-length v)) v (subu8vector v 0 o))]
        [else => (botan-error 'hex-decode s)]))))

(define (base64-encode v)
  (let*-values ([(v* vn) (data-object/length 'base64-encode v)]
                [(s) (make-string (add1 (ceiling (* 4/3 vn))))])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_base64_encode" (scheme-pointer (const unsigned-byte)) size_t (nonnull-scheme-pointer char) (c-pointer size_t))
             v* vn s (location n))
        [(0)  (substring s 0 (sub1 n))]
        [else => (botan-error 'base64-encode v)]))))

(define (base64-decode s)
  (let* ([n (string-length s)]
         [v (make-u8vector (ceiling (* 3/4 n)))])
    (let-location ([o size_t (u8vector-length v)])
      (case ((foreign-lambda int "botan_base64_decode" (nonnull-scheme-pointer (const char)) size_t nonnull-u8vector (c-pointer size_t))
             s n v (location o))
        [(0)  (if (= o (u8vector-length v)) v (subu8vector v 0 o))]
        [else => (botan-error 'base64-decode s)]))))

(define-pointer-type (botan-rng "botan_rng_struct")
  *botan-rng* rng?)

(define (rng-destroy! rng)
  (case ((foreign-lambda int "botan_rng_destroy" botan-rng) rng)
    [(0)  (void)]
    [else => (botan-error 'rng-destroy! rng)]))

(define (make-rng type)
  (when (symbol? type) (set! type (symbol->string type)))
  (let-location ([rng botan-rng #f])
    (case ((foreign-lambda int "botan_rng_init" (c-pointer botan-rng) nonnull-c-string) (location rng) type)
      [(0)  (set-finalizer! rng rng-destroy!)]
      [else => (botan-error 'make-rng type)])))

(define (rng-get rng v #!key [from 0] to)
  (when (integer? v) (set! v (make-u8vector v)))
  (set! to (if to (min (max 0 to) (u8vector-length v)) (u8vector-length v)))
  (set! from (min (max 0 from) to))
  (case ((foreign-lambda* int ([botan-rng rng] [nonnull-u8vector v] [size_t start] [size_t count])
           "C_return(botan_rng_get(rng, v + start, count));")
         rng v from (- to from))
    [(0)  v]
    [else => (botan-error 'rng-get rng v from: from to: to)]))

(define (rng-reseed! rng #!key [bits 256] source)
  (case (if source
          ((foreign-lambda int "botan_rng_reseed_from_rng" botan-rng botan-rng size_t) rng source bits)
          ((foreign-lambda int "botan_rng_reseed" botan-rng size_t) rng bits))
    [(0)  (void)]
    [else => (botan-error 'rng-reseed! rng bits: bits source: source)]))

(define (rng-add-entropy! rng v #!key [from 0] to)
  (receive (v* n) (data-object/length 'rng-add-entropy! v)
    (set! to (if to (min (max 0 to) n) n))
    (set! from (min (max 0 from) to))
    (case ((foreign-lambda* int ([botan-rng rng] [(scheme-pointer (const unsigned-byte)) v] [size_t start] [size_t count])
             "C_return(botan_rng_add_entropy(rng, v + start, count));")
           rng v* from (- to from))
      [(0)  (void)]
      [else => (botan-error 'rng-add-entropy! rng v from: from to: to)])))

(define-pointer-type (botan-hash "botan_hash_struct")
  *botan-hash* hash?)

(define (hash-destroy! hash)
  (case ((foreign-lambda int "botan_hash_destroy" botan-hash) hash)
    [(0)  (void)]
    [else => (botan-error 'hash-destroy! hash)]))

(define (make-hash name)
  (when (symbol? name) (set! name (symbol->string name)))
  (let-location ([hash botan-hash #f])
    (case ((foreign-lambda int "botan_hash_init" (c-pointer botan-hash) nonnull-c-string unsigned-int32)
           (location hash) name 0)
      [(0)  (set-finalizer! hash hash-destroy!)]
      [else => (botan-error 'make-hash name)])))

(define (hash-name hash)
  (let ([s (make-string 256)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_hash_name" botan-hash (nonnull-scheme-pointer char) (c-pointer size_t))
             hash s (location n))
        [(0)
         (set! s (substring s 0 (sub1 n)))
         (if (string-every char-set:nice-symbol s) (string->symbol s) s)]
        [else =>
         (botan-error 'hash-name hash)]))))

(define (hash-copy hash)
  (let-location ([hash1 botan-hash #f])
    (case ((foreign-lambda int "botan_hash_copy_state" (c-pointer botan-hash) botan-hash)
           (location hash1) hash)
      [(0)  (set-finalizer! hash1 hash-destroy!)]
      [else => (botan-error 'hash-copy hash)])))

(define (hash-clear! hash)
  (case ((foreign-lambda int "botan_hash_clear" botan-hash) hash)
    [(0)  (void)]
    [else => (botan-error 'hash-clear! hash)]))

(define (hash-update! hash data #!key [from 0] to)
  (receive (data* data-length) (data-object/length 'hash-update! data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (case ((foreign-lambda* int ([botan-hash hash] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count])
             "C_return(botan_hash_update(hash, data + start, count));")
           hash data* from data-length)
      [(0)  (void)]
      [else => (botan-error 'hash-update! hash data from: from to: to)])))

(define (hash-length hash)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_hash_output_length" botan-hash (c-pointer size_t))
           hash (location n))
      [(0)  n]
      [else => (botan-error 'hash-length hash)])))

(define (hash-final hash)
  (let ([v (make-u8vector (hash-length hash))])
    (case ((foreign-lambda int "botan_hash_final" botan-hash nonnull-u8vector) hash v)
      [(0)  v]
      [else => (botan-error 'hash-final hash)])))

(define-pointer-type (botan-mac "botan_mac_struct")
  *botan-mac* mac?)

(define (mac-destroy! mac)
  (case ((foreign-lambda int "botan_mac_destroy" botan-mac) mac)
    [(0)  (void)]
    [else => (botan-error 'mac-destroy! mac)]))

(define (make-mac name)
  (cond
    [(symbol? name)
     (set! name (symbol->string name))]
    [(and (pair? name) (pair? (cdr name)) (null? (cddr name)))
     (set! name (string-append (symbol->string (car name)) "(" (symbol->string (cadr name)) ")"))])
  (let-location ([mac botan-mac #f])
    (case ((foreign-lambda int "botan_mac_init" (c-pointer botan-mac) nonnull-c-string unsigned-int32)
           (location mac) name 0)
      [(0)  (set-finalizer! mac mac-destroy!)]
      [else => (botan-error 'make-mac name)])))

(define (mac-name mac)
  (let ([s (make-string 256)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_mac_name" botan-mac (nonnull-scheme-pointer char) (c-pointer size_t))
             mac s (location n))
        [(0)
         (set! s (substring s 0 (sub1 n)))
         (or
           (and-let* ([o (string-index s #\()]
                      [c (string-index s #\) o)]
                      [(= (string-length s) (add1 c))])
             (list (string->symbol (substring s 0 o)) (string->symbol (substring s (add1 o) c))))
           s)]
        [else =>
         (botan-error 'mac-name mac)]))))

(define (mac-clear! mac)
  (case ((foreign-lambda int "botan_mac_clear" botan-mac) mac)
    [(0)  (void)]
    [else => (botan-error 'mac-clear! mac)]))

(define (mac-keyspec mac)
  (let-location ([min size_t 0] [max size_t 0] [mod size_t 0])
    (case ((foreign-lambda int "botan_mac_get_keyspec" botan-mac (c-pointer size_t) (c-pointer size_t) (c-pointer size_t))
           mac (location min) (location max) (location mod))
      [(0)  (values min max mod)]
      [else => (botan-error 'mac-keyspec mac)])))

(define (mac-set-key! mac key)
  (case ((foreign-lambda int "botan_mac_set_key" botan-mac nonnull-u8vector size_t)
         mac key (u8vector-length key))
    [(0)  (void)]
    [else => (botan-error 'mac-set-key! mac)]))

(define (mac-update! mac data #!key [from 0] to)
  (receive (data* data-length) (data-object/length 'mac-update! data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (case ((foreign-lambda* int ([botan-mac mac] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count])
             "C_return(botan_mac_update(mac, data + start, count));")
           mac data* from data-length)
      [(0)  (void)]
      [else => (botan-error 'mac-update! mac data from: from to: to)])))

(define (mac-length mac)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_mac_output_length" botan-mac (c-pointer size_t))
           mac (location n))
      [(0)  n]
      [else => (botan-error 'mac-length mac)])))

(define (mac-final mac)
  (let ([v (make-u8vector (mac-length mac))])
    (case ((foreign-lambda int "botan_mac_final" botan-mac nonnull-u8vector) mac v)
      [(0)  v]
      [else => (botan-error 'mac-final mac)])))

(define-pointer-type (botan-cipher "botan_cipher_struct")
  *botan-cipher* cipher?)

(define (cipher-destroy! cipher)
  (case ((foreign-lambda int "botan_cipher_destroy" botan-cipher) cipher)
    [(0)  (void)]
    [else => (botan-error 'cipher-destroy! cipher)]))

(define (make-cipher name direction)
  (when (symbol? name) (set! name (symbol->string name)))
  (let-location ([cipher botan-cipher #f])
    (case ((foreign-lambda int "botan_cipher_init" (c-pointer botan-cipher) nonnull-c-string unsigned-int32)
           (location cipher) name
           (case direction
             [(encrypt) 0]
             [(decrypt) 1]
             [else      (type-error 'encrypt/decrypt direction)]))
      [(0)  (set-finalizer! cipher cipher-destroy!)]
      [else => (botan-error 'make-cipher name direction)])))

(define (cipher-name cipher)
  (let ([s (make-string 256)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_cipher_name" botan-cipher (nonnull-scheme-pointer char) (c-pointer size_t))
             cipher s (location n))
        [(0)
         (set! s (substring s 0 (sub1 n)))
         (if (string-every char-set:nice-symbol s) (string->symbol s) s)]
        [else => (botan-error 'cipher-name cipher)]))))

(define (cipher-clear! cipher)
  (case ((foreign-lambda int "botan_cipher_clear" botan-cipher) cipher)
    [(0)  (void)]
    [else => (botan-error 'cipher-clear! cipher)]))

(define (cipher-keyspec cipher)
  (let-location ([min size_t 0] [max size_t 0] [mod size_t 0])
    (case ((foreign-lambda int "botan_cipher_get_keyspec" botan-cipher (c-pointer size_t) (c-pointer size_t) (c-pointer size_t))
           cipher (location min) (location max) (location mod))
      [(0)  (values min max mod)]
      [else => (botan-error 'cipher-keyspec cipher)])))

(define (cipher-set-key! cipher key)
  (case ((foreign-lambda int "botan_cipher_set_key" botan-cipher nonnull-u8vector size_t)
         cipher key (u8vector-length key))
    [(0)  (void)]
    [else => (botan-error 'cipher-set-key! cipher)]))

(define (cipher-output-length cipher in)
  (let-location ([on size_t 0])
    (case ((foreign-lambda int "botan_cipher_output_length" botan-cipher size_t (c-pointer size_t))
           cipher in (location on))
      [(0)  on]
      [else => (botan-error 'cipher-output-length cipher in)])))

(define (cipher-tag-length cipher)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_cipher_get_tag_length" botan-cipher (c-pointer size_t))
           cipher (location n))
      [(0)  n]
      [else => (botan-error 'cipher-tag-length cipher)])))

(define (cipher-valid-nonce-length? cipher n)
  (case ((foreign-lambda int "botan_cipher_valid_nonce_length" botan-cipher size_t) cipher n)
    [(0)  #f]
    [(1)  #t]
    [else => (botan-error 'cipher-valid-nonce-length? cipher n)]))

(define (cipher-default-nonce-length cipher)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_cipher_get_default_nonce_length" botan-cipher (c-pointer size_t))
           cipher (location n))
      [(0)  n]
      [else => (botan-error 'cipher-default-nonce-length cipher)])))

(define (%cipher-reset! location cipher)
  (case ((foreign-lambda int "botan_cipher_reset" botan-cipher) cipher)
    [(0)  (void)]
    [else => (botan-error location cipher)]))

(define (%cipher-set-associated-data! location cipher data)
  (case (receive (data data-length) (data-object/length location data)
          ((foreign-lambda int "botan_cipher_set_associated_data" botan-cipher (scheme-pointer (const unsigned-byte)) size_t)
           cipher data data-length))
    [(0)  (void)]
    [else => (botan-error location cipher data)]))

(define (%cipher-start! location cipher nonce)
  (case ((foreign-lambda int "botan_cipher_start" botan-cipher nonnull-u8vector size_t)
         cipher nonce (u8vector-length nonce))
    [(0)  (void)]
    [else => (botan-error location cipher nonce)]))

#;
(define (%cipher-update-granularity location cipher)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_cipher_get_update_granularity" botan-cipher (c-pointer size_t))
           cipher (location n))
      [(0)  n]
      [else => (botan-error location cipher)])))

(define (%cipher-update! location cipher o ofrom oto i ifrom ito)
  (case ((foreign-lambda* int ([botan-cipher cipher]
                               [(nonnull-scheme-pointer unsigned-byte) o] [size_t ostart] [size_t ocount]
                               [(scheme-pointer (const unsigned-byte)) i] [size_t istart] [size_t icount])
           "C_return(botan_cipher_update(cipher, BOTAN_CIPHER_UPDATE_FLAG_FINAL, o + ostart, ocount, &ocount, i + istart, icount, &icount));")
         cipher o ofrom (- oto ofrom) i ifrom (- ito ifrom))
    [(0)  #t]
    [(-2) #f]
    [else => (botan-error location cipher)]))

(define (cipher-transform cipher data #!key [from 0] to aad nonce)
  (receive (ibuf ilen) (data-object/length 'cipher-transform data)
    (set! to (if to (min (max 0 to) ilen) ilen))
    (set! from (min (max 0 from) to))
    (let* ([olen (cipher-output-length cipher (- to from))]
           [obuf (if (string? data) (make-string olen) (make-blob olen))])
      (when aad (%cipher-set-associated-data! 'cipher-transform cipher aad))
      (when nonce (%cipher-start! 'cipher-transform cipher nonce))
      (let ([ok (%cipher-update! 'cipher-transform cipher obuf 0 olen ibuf from to)])
        (%cipher-reset! 'cipher-transform cipher)
        (and ok (if (u8vector? data) (blob->u8vector/shared obuf) obuf))))))

(define pbkdf-iterations
  (make-parameter 10000))

(define (pbkdf algo v password #!key [from 0] to salt [iterations (pbkdf-iterations)] milliseconds)
  (cond
    [(symbol? algo)
     (set! algo (symbol->string algo))]
    [(and (pair? algo) (pair? (cdr algo)) (null? (cddr algo)))
     (set! algo (string-append (symbol->string (car algo)) "(" (symbol->string (cadr algo)) ")"))])
  (receive (salt* salt-length) (data-object/length 'pbkdf salt)
    (when (integer? v) (set! v (make-u8vector v)))
    (set! to (if to (min (max 0 to) (u8vector-length v)) (u8vector-length v)))
    (set! from (min (max 0 from) to))
    (let-location ([iterations* size_t iterations])
      (case (if milliseconds
              ((foreign-lambda* int ([nonnull-c-string algo] [nonnull-u8vector o] [size_t start] [size_t count]
                                     [(scheme-pointer (const char)) password] [size_t npassword]
                                     [(scheme-pointer (const unsigned-byte)) salt] [size_t nsalt]
                                     [unsigned-int32 milliseconds] [(c-pointer size_t) iterations])
                 "C_return(botan_pwdhash_timed(algo, milliseconds, iterations, NULL, NULL, o + start, count, password, npassword, salt, nsalt));")
               algo v from (- to from)
               password (string-length password) salt* salt-length
               milliseconds (location iterations*))
              ((foreign-lambda* int ([nonnull-c-string algo] [nonnull-u8vector o] [size_t start] [size_t count]
                                     [(scheme-pointer (const char)) password] [size_t npassword]
                                     [(scheme-pointer (const unsigned-byte)) salt] [size_t nsalt]
                                     [size_t iterations])
                 "C_return(botan_pwdhash(algo, iterations, 0, 0, o + start, count, password, npassword, salt, nsalt));")
               algo v from (- to from)
               password (string-length password) salt* salt-length
               iterations))
        [(0)
         (pbkdf-iterations iterations*)
         v]
        [else =>
         (botan-error 'pbkdf algo v from: from to: to salt: salt iterations: iterations milliseconds: milliseconds)]))))

(define (kdf algo v secret #!key [from 0] to salt label)
  (cond
    [(symbol? algo)
     (set! algo (symbol->string algo))]
    [(and (pair? algo) (pair? (cdr algo)) (null? (cddr algo)))
     (set! algo (string-append (symbol->string (car algo)) "(" (symbol->string (cadr algo)) ")"))])
  (let-values ([(secret* secret-length) (data-object/length 'kdf secret)]
               [(salt* salt-length) (data-object/length 'kdf salt)])
    (when (integer? v) (set! v (make-u8vector v)))
    (set! to (if to (min (max 0 to) (u8vector-length v)) (u8vector-length v)))
    (set! from (min (max 0 from) to))
    (case ((foreign-lambda* int ([nonnull-c-string algo] [nonnull-u8vector o] [size_t start] [size_t count]
                                 [(scheme-pointer (const unsigned-byte)) secret] [size_t nsecret]
                                 [(scheme-pointer (const unsigned-byte)) salt] [size_t nsalt]
                                 [(scheme-pointer (const unsigned-byte)) label] [size_t nlabel])
             "C_return(botan_kdf(algo, o + start, count, secret, nsecret, salt, nsalt, label, nlabel));")
           algo v from (- to from)
           secret* secret-length salt* salt-length label (string-length label))
      [(0)  v]
      [else => (botan-error 'kdf algo v from: from to: to salt: salt label: label)])))

(define (bcrypt-generate password rng #!key [work-factor 10])
  (let ([hash (make-string 64)])
    (let-location ([n size_t (string-length hash)])
      (case ((foreign-lambda int "botan_bcrypt_generate" (nonnull-scheme-pointer unsigned-byte) (c-pointer size_t) c-string botan-rng size_t unsigned-int32)
             hash (location n) password rng work-factor 0)
        [(0)  (substring hash 0 (sub1 n))]
        [else => (botan-error 'bcrypt-generate rng work-factor: work-factor)]))))

(define (bcrypt-valid? password hash)
  (case ((foreign-lambda int "botan_bcrypt_is_valid" c-string c-string) password hash)
    [(0)  #t]
    [(1)  #f]
    [else => (botan-error 'bcrypt-valid? hash)]))

(define-pointer-type (botan-mp "botan_mp_struct")
  *botan-mp* mp?)

(define (mp-destroy! mp)
  (case ((foreign-lambda int "botan_mp_destroy" botan-mp) mp)
    [(0)  (void)]
    [else => (botan-error 'mp-destroy! mp)]))

(define (%make-mp caller)
  (let-location ([mp botan-mp #f])
    (case ((foreign-lambda int "botan_mp_init" (c-pointer botan-mp)) (location mp))
      [(0)  (set-finalizer! mp mp-destroy!)]
      [else => (botan-error caller)])))

(define (%mp-bytes caller mp)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_mp_num_bytes" botan-mp (c-pointer size_t)) mp (location n))
      [(0)  n]
      [else => (botan-error caller mp)])))

(define (%mp->hex caller mp)
  (let ([s (make-string (add1 (* 2 (%mp-bytes caller mp))))])
    (case ((foreign-lambda int "botan_mp_to_hex" botan-mp (scheme-pointer char)) mp s)
      [(0)  (substring s 0 (sub1 (string-length s)))]
      [else => (botan-error caller mp)])))

(define (%mp->number caller mp)
  (string->number (%mp->hex caller mp) 16))

(define (%number->mp caller n)
  (let-location ([mp botan-mp #f])
    (case ((foreign-lambda int "botan_mp_init" (c-pointer botan-mp)) (location mp))
      [(0)
       (let ([mp (set-finalizer! mp mp-destroy!)])
	 (cond
	  [(zero? n)
	   mp]
	  [(fixnum? n)
	   (case ((foreign-lambda int "botan_mp_set_from_int" botan-mp int) mp n)
	     [(0)  mp]
	     [else => (botan-error caller n)])]
	  [else
	   (let ([s (string-append "0x" (number->string (abs n)) "\x00")])
	     (case ((foreign-lambda int "botan_mp_set_from_str" botan-mp scheme-pointer) mp s)
	       [(0)
		(if (negative? n)
		  (case ((foreign-lambda int "botan_mp_flip_sign" botan-mp) mp)
		    [(0)  mp]
		    [else => (botan-error caller n)])
		  mp)]
	       [else =>
		     (botan-error caller n)]))]))]
      [else =>
       (botan-error caller n)])))

(define-pointer-type (botan-privkey "botan_privkey_struct")
  *botan-privkey* privkey?)

(define (privkey-destroy! sk)
  (case ((foreign-lambda int "botan_privkey_destroy" botan-privkey) sk)
    [(0)  (void)]
    [else => (botan-error 'privkey-destroy! sk)]))

(define (generate-privkey algo rng #!optional params)
  (when (symbol? algo) (set! algo (symbol->string algo)))
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_create" (c-pointer botan-privkey) nonnull-c-string c-string botan-rng)
           (location sk) algo params rng)
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'generate-privkey algo rng params)])))

(define (make-privkey/rsa #!key p q e)
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_rsa" (c-pointer botan-privkey) botan-mp botan-mp botan-mp)
	   (location sk) (%number->mp 'make-privkey/rsa p) (%number->mp 'make-privkey/rsa q) (%number->mp 'make-privkey/rsa e))
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/rsa)])))

(define (make-privkey/dsa #!key p q g x)
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_dsa" (c-pointer botan-privkey) botan-mp botan-mp botan-mp botan-mp)
	   (location sk) (%number->mp 'make-privkey/dsa p) (%number->mp 'make-privkey/dsa q) (%number->mp 'make-privkey/dsa g) (%number->mp 'make-privkey/dsa x))
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/dsa)])))

(define (make-privkey/elgamal #!key p g x)
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_elgamal" (c-pointer botan-privkey) botan-mp botan-mp botan-mp)
	   (location sk) (%number->mp 'make-privkey/elgamal p) (%number->mp 'make-privkey/elgamal g) (%number->mp 'make-privkey/elgamal x))
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/elgamal)])))

(define (make-privkey/dh #!key p g x)
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_dh" (c-pointer botan-privkey) botan-mp botan-mp botan-mp)
	   (location sk) (%number->mp 'make-privkey/dh p) (%number->mp 'make-privkey/dh g) (%number->mp 'make-privkey/dh x))
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/dh)])))

(define (make-privkey/ed25519 v)
  (unless (= (u8vector-length v) 64)
    (type-error 'make-privkey/ed25519 v))
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_ed25519" (c-pointer botan-privkey) nonnull-u8vector)
	   (location sk) v)
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/ed25519)])))

(define (make-privkey/x25519 v)
  (unless (= (u8vector-length v) 32)
    (type-error 'make-privkey/x25519 v))
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_x25519" (c-pointer botan-privkey) nonnull-u8vector)
	   (location sk) v)
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/x25519)])))

(define (make-privkey/ecdsa #!key scalar curve)
  (when (symbol? curve) (set! curve (symbol->string curve)))
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_ecdsa" (c-pointer botan-privkey) botan-mp nonnull-c-string)
	   (location sk) (%number->mp 'make-privkey/ecdsa scalar) curve)
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/ecdsa curve: curve)])))

(define (make-privkey/ecdh #!key scalar curve)
  (when (symbol? curve) (set! curve (symbol->string curve)))
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_ecdh" (c-pointer botan-privkey) botan-mp nonnull-c-string)
	   (location sk) (%number->mp 'make-privkey/ecdh scalar) curve)
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/ecdh curve: curve)])))

(define (make-privkey/sm2 #!key scalar curve)
  (when (symbol? curve) (set! curve (symbol->string curve)))
  (let-location ([sk botan-privkey #f])
    (case ((foreign-lambda int "botan_privkey_load_sm2" (c-pointer botan-privkey) botan-mp nonnull-c-string)
	   (location sk) (%number->mp 'make-privkey/sm2 scalar) curve)
      [(0)  (set-finalizer! sk privkey-destroy!)]
      [else => (botan-error 'make-privkey/sm2 curve: curve)])))

(define (data->privkey data #!key [from 0] to password)
  (receive (data* data-length) (data-object/length 'data->privkey data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (let-location ([sk botan-privkey #f])
      (case ((foreign-lambda* int ([(c-pointer botan-privkey) sk] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count] [c-string password])
               "C_return(botan_privkey_load(sk, NULL, data + start, count, password));")
             (location sk) data* from data-length password)
        [(0)  (set-finalizer! sk privkey-destroy!)]
        [else => (botan-error 'data->privkey data from: from to: to)]))))

(define (load-privkey path #!key password)
  (data->privkey
   (call-with-input-file path (cut read-string #f <>))
   password: password))

(define (%privkey-export! caller sk v flags)
  (let-location ([n size_t (if (string? v) (string-length v) (blob-size v))])
    (case ((foreign-lambda int "botan_privkey_export" botan-privkey (scheme-pointer unsigned-byte) (c-pointer size_t) unsigned-int32)
           sk v (location n) flags)
      [(0)  n]
      [else => (botan-error caller sk)])))

(define (%privkey-export/pbkdf-timed! caller sk v rng password milliseconds cipher hash flags)
  (when (symbol? cipher) (set! cipher (symbol->string cipher)))
  (when (symbol? hash) (set! hash (symbol->string hash)))
  (let-location ([n size_t (if (string? v) (string-length v) (blob-size v))]
                 [iterations* size_t (pbkdf-iterations)])
    (case ((foreign-lambda int "botan_privkey_export_encrypted_pbkdf_msec" botan-privkey (scheme-pointer unsigned-byte) (c-pointer size_t) botan-rng c-string unsigned-int32 (c-pointer size_t) nonnull-c-string nonnull-c-string unsigned-int32)
           sk v (location n) rng password milliseconds (location iterations*) cipher hash flags)
      [(0)
       (pbkdf-iterations iterations*)
       n]
      [else =>
       (botan-error caller sk)])))

(define (%privkey-export/pbkdf! caller sk v rng password iterations cipher hash flags)
  (when (symbol? cipher) (set! cipher (symbol->string cipher)))
  (when (symbol? hash) (set! hash (symbol->string hash)))
  (let-location ([n size_t (if (string? v) (string-length v) (blob-size v))])
    (case ((foreign-lambda int "botan_privkey_export_encrypted_pbkdf_iter" botan-privkey (scheme-pointer unsigned-byte) (c-pointer size_t) botan-rng c-string size_t nonnull-c-string nonnull-c-string unsigned-int32)
           sk v (location n) rng password iterations cipher hash flags)
      [(0)  n]
      [else => (botan-error caller sk)])))

(define (privkey->string sk #!key rng password [cipher 'AES-256/GCM] [hash 'SHA-256] [iterations (pbkdf-iterations)] milliseconds)
  (let* ([s (make-string (+ (* 16 (pubkey-estimated-strength (privkey->pubkey sk))) 1024))]
         [n (cond
              [(not password)
               (%privkey-export! 'privkey->string sk s 1)]
              [(not milliseconds)
               (%privkey-export/pbkdf! 'privkey->string sk s rng password iterations cipher hash 1)]
              [else
               (%privkey-export/pbkdf-timed! 'privkey->string sk s rng password milliseconds cipher hash 1)])])
    (substring s 0 (sub1 n))))

(define (privkey->blob sk #!key rng password [cipher 'AES-256/GCM] [hash 'SHA-256] [iterations 10000] milliseconds)
  (let* ([v (make-u8vector (+ (* 8 (pubkey-estimated-strength (privkey->pubkey sk))) 1024))]
         [n (cond
              [(not password)
               (%privkey-export! 'privkey->string sk (u8vector->blob/shared v) 0)]
              [(not milliseconds)
               (%privkey-export/pbkdf! 'privkey->string sk (u8vector->blob/shared v) rng password iterations cipher hash 0)]
              [else
               (%privkey-export/pbkdf-timed! 'privkey->string sk (u8vector->blob/shared v) rng password milliseconds cipher hash 0)])])
    (u8vector->blob/shared (subu8vector v 0 n))))

(define (privkey->ed25519 sk)
  (let ([v (make-u8vector 64)])
    (case ((foreign-lambda int "botan_privkey_ed25519_get_privkey" botan-privkey nonnull-u8vector)
	   sk v)
      [(0)  v]
      [else => (botan-error 'privkey->ed25519 sk)])))

(define (privkey->x25519 sk)
  (let ([v (make-u8vector 64)])
    (case ((foreign-lambda int "botan_privkey_x25519_get_privkey" botan-privkey nonnull-u8vector)
	   sk v)
      [(0)  v]
      [else => (botan-error 'privkey->x25519 sk)])))

(define (privkey-check sk rng #!key strong)
  (case ((foreign-lambda int "botan_privkey_check_key" botan-privkey botan-rng unsigned-int32)
         sk rng (if strong 1 0))
    [(0)  #t]
    [(-1) #f]
    [else => (botan-error 'privkey-check sk rng strong: strong)]))

(define (privkey-algorithm sk)
  (let ([s (make-string 256)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_privkey_algo_name" botan-privkey (nonnull-scheme-pointer char) (c-pointer size_t))
             sk s (location n))
        [(0)
         (set! s (substring s 0 (sub1 n)))
         (if (string-every char-set:nice-symbol s) (string->symbol s) s)]
        [else =>
         (botan-error 'privkey-algorithm sk)]))))

(define (privkey-field sk field)
  (let ([mp (%make-mp 'privkey-field)])
    (case ((foreign-lambda int "botan_privkey_get_field" botan-mp botan-privkey symbol) mp sk field)
      [(0)  (%mp->number 'privkey-field mp)]
      [else => (botan-error 'privkey-field sk field)])))

(define-pointer-type (botan-pubkey "botan_pubkey_struct")
  *botan-pubkey* pubkey?)

(define (pubkey-destroy! pk)
  (case ((foreign-lambda int "botan_pubkey_destroy" botan-pubkey) pk)
    [(0)  (void)]
    [else => (botan-error 'pubkey-destroy! pk)]))

(define (privkey->pubkey sk)
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_privkey_export_pubkey" (c-pointer botan-pubkey) botan-privkey)
           (location pk) sk)
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'privkey->pubkey sk)])))

(define (make-pubkey/rsa #!key n e)
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_rsa" (c-pointer botan-pubkey) botan-mp botan-mp)
	   (location pk) (%number->mp 'make-pubkey/rsa n) (%number->mp 'make-pubkey/rsa e))
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/rsa n: n e: e)])))

(define (make-pubkey/dsa #!key p q g y)
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_dsa" (c-pointer botan-pubkey) botan-mp botan-mp botan-mp botan-mp)
	   (location pk) (%number->mp 'make-pubkey/dsa p) (%number->mp 'make-pubkey/dsa q) (%number->mp 'make-pubkey/dsa g) (%number->mp 'make-pubkey/dsa y))
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/dsa p: p q: q g: g y: y)])))

(define (make-pubkey/elgamal #!key p g y)
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_elgamal" (c-pointer botan-pubkey) botan-mp botan-mp botan-mp)
	   (location pk) (%number->mp 'make-pubkey/elgamal p) (%number->mp 'make-pubkey/elgamal g) (%number->mp 'make-pubkey/elgamal y))
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/elgamal p: p g: g y: y)])))

(define (make-pubkey/dh #!key p g y)
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_dh" (c-pointer botan-pubkey) botan-mp botan-mp botan-mp)
	   (location pk) (%number->mp 'make-pubkey/dh p) (%number->mp 'make-pubkey/dh g) (%number->mp 'make-pubkey/dh y))
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/dh p: p g: g y: y)])))

(define (make-pubkey/ed25519 v)
  (unless (= (u8vector-length v) 32)
    (type-error 'make-pubkey/ed25519 v))
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_ed25519" (c-pointer botan-pubkey) nonnull-u8vector)
	   (location pk) v)
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/ed25519 v)])))

(define (make-pubkey/x25519 v)
  (unless (= (u8vector-length v) 32)
    (type-error 'make-pubkey/x25519 v))
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_x25519" (c-pointer botan-pubkey) nonnull-u8vector)
	   (location pk) v)
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/x25519 v)])))

(define (make-pubkey/ecdsa #!key x y curve)
  (when (symbol? curve) (set! curve (symbol->string curve)))
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_ecdsa" (c-pointer botan-pubkey) botan-mp botan-mp nonnull-c-string)
	   (location pk) (%number->mp 'make-pubkey/ecdsa x) (%number->mp 'make-pubkey/ecdsa y) curve)
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/ecdsa x: x y: y curve: curve)])))

(define (make-pubkey/ecdh #!key x y curve)
  (when (symbol? curve) (set! curve (symbol->string curve)))
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_ecdh" (c-pointer botan-pubkey) botan-mp botan-mp nonnull-c-string)
	   (location pk) (%number->mp 'make-pubkey/ecdh x) (%number->mp 'make-pubkey/ecdh y) curve)
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/ecdh x: x y: y curve: curve)])))

(define (make-pubkey/sm2 #!key x y curve)
  (when (symbol? curve) (set! curve (symbol->string curve)))
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_pubkey_load_sm2" (c-pointer botan-pubkey) botan-mp botan-mp nonnull-c-string)
	   (location pk) (%number->mp 'make-pubkey/sm2 x) (%number->mp 'make-pubkey/sm2 y) curve)
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'make-pubkey/sm2 x: x y: y curve: curve)])))

(define (data->pubkey data #!key [from 0] to)
  (receive (data* data-length) (data-object/length 'data->pubkey data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (let-location ([pk botan-pubkey #f])
      (case ((foreign-lambda* int ([(c-pointer botan-pubkey) pk] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count])
               "C_return(botan_pubkey_load(pk, data + start, count));")
             (location pk) data* from data-length)
        [(0)  (set-finalizer! pk pubkey-destroy!)]
        [else => (botan-error 'data->pubkey data from: from to: to)]))))

(define (load-pubkey path)
  (data->pubkey
   (call-with-input-file path (cut read-string #f <>))))

(define (%pubkey-export! caller pk v flags)
  (let-location ([n size_t (if (string? v) (string-length v) (blob-size v))])
    (case ((foreign-lambda int "botan_pubkey_export" botan-pubkey (scheme-pointer unsigned-byte) (c-pointer size_t) unsigned-int32)
           pk v (location n) flags)
      [(0)  n]
      [else => (botan-error caller pk)])))

(define (pubkey->string pk)
  (let* ([s (make-string (+ (* 16 (pubkey-estimated-strength pk)) 1024))]
         [n (%pubkey-export! 'pubkey->string pk s 1)])
    (substring s 0 (sub1 n))))

(define (pubkey->blob pk)
  (let* ([v (make-u8vector (+ (* 8 (pubkey-estimated-strength pk)) 1024))]
         [n (%pubkey-export! 'pubkey->string pk (u8vector->blob/shared v) 0)])
    (u8vector->blob/shared (subu8vector v 0 n))))

(define (pubkey->ed25519 pk)
  (let ([v (make-u8vector 64)])
    (case ((foreign-lambda int "botan_pubkey_ed25519_get_pubkey" botan-pubkey nonnull-u8vector)
	   pk v)
      [(0)  v]
      [else => (botan-error 'pubkey->ed25519 pk)])))

(define (pubkey->x25519 pk)
  (let ([v (make-u8vector 64)])
    (case ((foreign-lambda int "botan_pubkey_x25519_get_pubkey" botan-pubkey nonnull-u8vector)
	   pk v)
      [(0)  v]
      [else => (botan-error 'pubkey->x25519 pk)])))

(define (pubkey-algorithm pk)
  (let ([s (make-string 256)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_pubkey_algo_name" botan-pubkey (nonnull-scheme-pointer char) (c-pointer size_t))
             pk s (location n))
        [(0)
         (set! s (substring s 0 (sub1 n)))
         (if (string-every char-set:nice-symbol s) (string->symbol s) s)]
        [else =>
         (botan-error 'pubkey-algorithm pk)]))))

(define (pubkey-field pk field)
  (let ([mp (%make-mp 'pubkey-field)])
    (case ((foreign-lambda int "botan_pubkey_get_field" botan-mp botan-pubkey symbol) mp pk field)
      [(0)  (%mp->number 'pubkey-field mp)]
      [else => (botan-error 'pubkey-field pk field)])))

(define (pubkey-estimated-strength pk)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_pubkey_estimated_strength" botan-pubkey (c-pointer size_t))
           pk (location n))
      [(0)  n]
      [else => (botan-error 'pubkey-estimated-strength pk)])))

(define (pubkey-fingerprint pk #!key [hash 'SHA-256])
  (when (symbol? hash) (set! hash (symbol->string hash)))
  (let ([v (make-u8vector 256)])
    (let-location ([n size_t (u8vector-length v)])
      (case ((foreign-lambda int "botan_pubkey_fingerprint" botan-pubkey nonnull-c-string nonnull-u8vector (c-pointer size_t))
             pk hash v (location n))
        [(0)  (if (= n (u8vector-length v)) v (subu8vector v 0 n))]
        [else => (botan-error 'pubkey-fingerprint pk hash: hash)]))))

(define-pointer-type (botan-pk-encrypt "botan_pk_op_encrypt_struct")
  *botan-pk-encrypt* pk-encrypt?)

(define (pk-encrypt-destroy! op)
  (case ((foreign-lambda int "botan_pk_op_encrypt_destroy" botan-pk-encrypt) op)
    [(0)  (void)]
    [else => (botan-error 'pk-encrypt-destroy! op)]))

(define (make-pk-encrypt pk padding)
  (cond
    [(symbol? padding)
     (set! padding (symbol->string padding))]
    [(and (pair? padding) (pair? (cdr padding)) (null? (cddr padding)))
     (set! padding (string-append (symbol->string (car padding)) "(" (symbol->string (cadr padding)) ")"))])
  (let-location ([op botan-pk-encrypt #f])
    (case ((foreign-lambda int "botan_pk_op_encrypt_create" (c-pointer botan-pk-encrypt) botan-pubkey nonnull-c-string unsigned-int32)
           (location op) pk padding 0)
      [(0)  (set-finalizer! op pk-encrypt-destroy!)]
      [else => (botan-error 'make-pk-encrypt pk padding)])))

(define (pk-encrypt-output-length op pt-length)
  (let-location ([ct-length size_t 0])
    (case ((foreign-lambda int "botan_pk_op_encrypt_output_length" botan-pk-encrypt size_t (c-pointer size_t))
           op pt-length (location ct-length))
      [(0)  ct-length]
      [else => (botan-error 'pk-encrypt-output-length op pt-length)])))

(define (pk-encrypt op rng pt #!key [from 0] to)
  (receive (pt* pt-length) (data-object/length 'pk-encrypt pt)
    (set! to (if to (min (max 0 to) pt-length) pt-length))
    (set! from (min (max 0 from) to))
    (set! pt-length (- to from))
    (let* ([ct-length (pk-encrypt-output-length op pt-length)]
           [ct (if (string? pt) (make-string ct-length) (make-blob ct-length))])
      (let-location ([n size_t ct-length])
        (case ((foreign-lambda* int ([botan-pk-encrypt op] [botan-rng rng]
                                     [(scheme-pointer unsigned-byte) ct] [(c-pointer size_t) ct_length]
                                     [(scheme-pointer (const unsigned-byte)) pt] [size_t start] [size_t pt_length])
                 "C_return(botan_pk_op_encrypt(op, rng, ct, ct_length, pt + start, pt_length));")
               op rng ct (location n) pt* from pt-length)
          [(0)
           (cond
             [(= n ct-length) (if (u8vector? pt) (blob->u8vector/shared ct) ct)]
             [(string? pt) (substring ct 0 n)]
             [(blob? pt) (u8vector->blob/shared (subu8vector (blob->u8vector/shared ct) 0 n))]
             [else (subu8vector (blob->u8vector/shared ct) 0 n)])]
          [else =>
           (botan-error 'pk-encrypt op rng pt from: from to: to)])))))

(define-pointer-type (botan-pk-decrypt "botan_pk_op_decrypt_struct")
  *botan-pk-decrypt* pk-decrypt?)

(define (pk-decrypt-destroy! op)
  (case ((foreign-lambda int "botan_pk_op_decrypt_destroy" botan-pk-decrypt) op)
    [(0)  (void)]
    [else => (botan-error 'pk-decrypt-destroy! op)]))

(define (make-pk-decrypt sk padding)
  (cond
    [(symbol? padding)
     (set! padding (symbol->string padding))]
    [(and (pair? padding) (pair? (cdr padding)) (null? (cddr padding)))
     (set! padding (string-append (symbol->string (car padding)) "(" (symbol->string (cadr padding)) ")"))])
  (let-location ([op botan-pk-decrypt #f])
    (case ((foreign-lambda int "botan_pk_op_decrypt_create" (c-pointer botan-pk-decrypt) botan-privkey nonnull-c-string unsigned-int32)
           (location op) sk padding 0)
      [(0)  (set-finalizer! op pk-decrypt-destroy!)]
      [else => (botan-error 'make-pk-decrypt sk padding)])))

(define (pk-decrypt-output-length op ct-length)
  (let-location ([pt-length size_t 0])
    (case ((foreign-lambda int "botan_pk_op_decrypt_output_length" botan-pk-decrypt size_t (c-pointer size_t))
           op ct-length (location pt-length))
      [(0)  pt-length]
      [else => (botan-error 'pk-decrypt-output-length op pt-length)])))

(define (pk-decrypt op ct #!key [from 0] to)
  (receive (ct* ct-length) (data-object/length 'pk-decrypt ct)
    (set! to (if to (min (max 0 to) ct-length) ct-length))
    (set! from (min (max 0 from) to))
    (set! ct-length (- to from))
    (let* ([pt-length (pk-decrypt-output-length op ct-length)]
           [pt (if (string? ct) (make-string pt-length) (make-blob pt-length))])
      (let-location ([n size_t pt-length])
        (case ((foreign-lambda* int ([botan-pk-decrypt op]
                                     [(scheme-pointer unsigned-byte) pt] [(c-pointer size_t) pt_length]
                                     [(scheme-pointer (const unsigned-byte)) ct] [size_t start] [size_t ct_length])
                 "C_return(botan_pk_op_decrypt(op, pt, pt_length, ct + start, ct_length));")
               op pt (location n) ct* from ct-length)
          [(0)
           (cond
             [(= n pt-length) (if (u8vector? ct) (blob->u8vector/shared pt) pt)]
             [(string? ct) (substring pt 0 n)]
             [(blob? ct) (u8vector->blob/shared (subu8vector (blob->u8vector/shared pt) 0 n))]
             [else (subu8vector (blob->u8vector/shared pt) 0 n)])]
          [(-1)
           #f]
          [else =>
           (botan-error 'pk-decrypt op ct from: from to: to)])))))

(define-pointer-type (botan-pk-sign "botan_pk_op_sign_struct")
  *botan-pk-sign* pk-sign?)

(define (pk-sign-destroy! op)
  (case ((foreign-lambda int "botan_pk_op_sign_destroy" botan-pk-sign) op)
    [(0)  (void)]
    [else => (botan-error 'pk-sign-destroy! op)]))

(define (make-pk-sign sk padding)
  (cond
    [(symbol? padding)
     (set! padding (symbol->string padding))]
    [(and (pair? padding) (pair? (cdr padding)) (null? (cddr padding)))
     (set! padding (string-append (symbol->string (car padding)) "(" (symbol->string (cadr padding)) ")"))])
  (let-location ([op botan-pk-sign #f])
    (case ((foreign-lambda int "botan_pk_op_sign_create" (c-pointer botan-pk-sign) botan-privkey nonnull-c-string unsigned-int32)
           (location op) sk padding 0)
      [(0)  (set-finalizer! op pk-sign-destroy!)]
      [else => (botan-error 'make-pk-sign sk padding)])))

(define (pk-sign-length op)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_pk_op_sign_output_length" botan-pk-sign (c-pointer size_t))
           op (location n))
      [(0)  n]
      [else => (botan-error 'pk-sign-length op)])))

(define (pk-sign-update! op data #!key [from 0] to)
  (receive (data* data-length) (data-object/length 'pk-sign-update! data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (case ((foreign-lambda* int ([botan-pk-sign op] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count])
             "C_return(botan_pk_op_sign_update(op, data + start, count));")
           op data* from data-length)
      [(0)  (void)]
      [else => (botan-error 'pk-sign-update! op data from: from to: to)])))

(define (pk-sign-final op rng)
  (let ([v (make-u8vector (pk-sign-length op))])
    (let-location ([n size_t (u8vector-length v)])
      (case ((foreign-lambda int "botan_pk_op_sign_finish" botan-pk-sign botan-rng nonnull-u8vector (c-pointer size_t))
             op rng v (location n))
        [(0)  (if (= n (u8vector-length v)) v (subu8vector v 0 n))]
        [else => (botan-error 'pk-sign-final op rng)]))))

(define-pointer-type (botan-pk-verify "botan_pk_op_verify_struct")
  *botan-pk-verify* pk-verify?)

(define (pk-verify-destroy! op)
  (case ((foreign-lambda int "botan_pk_op_verify_destroy" botan-pk-verify) op)
    [(0)  (void)]
    [else => (botan-error 'pk-verify-destroy! op)]))

(define (make-pk-verify pk padding)
  (cond
    [(symbol? padding)
     (set! padding (symbol->string padding))]
    [(and (pair? padding) (pair? (cdr padding)) (null? (cddr padding)))
     (set! padding (string-append (symbol->string (car padding)) "(" (symbol->string (cadr padding)) ")"))])
  (let-location ([op botan-pk-verify #f])
    (case ((foreign-lambda int "botan_pk_op_verify_create" (c-pointer botan-pk-verify) botan-pubkey nonnull-c-string unsigned-int32)
           (location op) pk padding 0)
      [(0)  (set-finalizer! op pk-verify-destroy!)]
      [else => (botan-error 'make-pk-verify pk padding)])))

(define (pk-verify-update! op data #!key [from 0] to)
  (receive (data* data-length) (data-object/length 'pk-verify-update! data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (case ((foreign-lambda* int ([botan-pk-verify op] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count])
             "C_return(botan_pk_op_verify_update(op, data + start, count));")
           op data* from data-length)
      [(0)  (void)]
      [else => (botan-error 'pk-verify-update! op data from: from to: to)])))

(define (pk-verify-final op data #!key [from 0] to)
  (receive (data* data-length) (data-object/length 'pk-verify-final data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (case ((foreign-lambda* int ([botan-pk-verify op] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count])
             "C_return(botan_pk_op_verify_finish(op, data + start, count));")
           op data* from data-length)
      [(0)  #t]
      [(1)  #f]
      [else => (botan-error 'pk-verify-final op data from: from to: to)])))

(define-pointer-type (botan-pk-key-agreement "botan_pk_op_ka_struct")
  *botan-pk-key-agreement* pk-key-agreement?)

(define (pk-key-agreement-destroy! op)
  (case ((foreign-lambda int "botan_pk_op_key_agreement_destroy" botan-pk-key-agreement) op)
    [(0)  (void)]
    [else => (botan-error 'pk-key-agreement-destroy! op)]))

(define (make-pk-key-agreement sk kdf)
  (cond
    [(symbol? kdf)
     (set! kdf (symbol->string kdf))]
    [(and (pair? kdf) (pair? (cdr kdf)) (null? (cddr kdf)))
     (set! kdf (string-append (symbol->string (car kdf)) "(" (symbol->string (cadr kdf)) ")"))])
  (let-location ([op botan-pk-key-agreement #f])
    (case ((foreign-lambda int "botan_pk_op_key_agreement_create" (c-pointer botan-pk-key-agreement) botan-privkey nonnull-c-string unsigned-int32)
           (location op) sk kdf 0)
      [(0)  (set-finalizer! op pk-key-agreement-destroy!)]
      [else => (botan-error 'make-pk-key-agreement sk kdf)])))

(define (pk-key-agreement-length op)
  (let-location ([n size_t 0])
    (case ((foreign-lambda int "botan_pk_op_key_agreement_size" botan-pk-key-agreement (c-pointer size_t))
           op (location n))
      [(0)  n]
      [else => (botan-error 'pk-key-agreement-length op)])))

(define (pk-key-agreement-export-public sk)
  (let ([v (make-u8vector (+ (* 8 (pubkey-estimated-strength (privkey->pubkey sk))) 1024))])
    (let-location ([n size_t (u8vector-length v)])
      (case ((foreign-lambda int "botan_pk_op_key_agreement_export_public" botan-privkey nonnull-u8vector (c-pointer size_t))
             sk v (location n))
        [(0)  (subu8vector v 0 n)]
        [else => (botan-error 'pk-key-agreement-export-public sk)]))))

(define (pk-key-agreement op other #!key salt)
  (let-values ([(salt* salt-length) (data-object/length 'pk-key-agreement salt)]
               [(v) (make-u8vector (pk-key-agreement-length op))])
    (let-location ([n size_t (u8vector-length v)])
      (case ((foreign-lambda* int ([botan-pk-key-agreement op] [nonnull-u8vector v] [(c-pointer size_t) count]
                                   [nonnull-u8vector other] [size_t nother]
                                   [(scheme-pointer (const unsigned-byte)) salt] [size_t nsalt])
               "C_return(botan_pk_op_key_agreement(op, v, count, other, nother, salt, nsalt));")
             op v (location n) other (u8vector-length other) salt* salt-length)
        [(0)  (if (= n (u8vector-length v)) v (subu8vector v 0 n))]
        [else => (botan-error 'pk-key-agreement op other salt: salt)]))))

(define-pointer-type (botan-x509-cert "botan_x509_cert_struct")
  *botan-x509-cert* x509-cert?)

(define (x509-cert-destroy! cert)
  (case ((foreign-lambda int "botan_x509_cert_destroy" botan-x509-cert) cert)
    [(0)  (void)]
    [else => (botan-error 'x509-cert-destroy! cert)]))

(define (data->x509-cert data #!key [from 0] to)
  (receive (data* data-length) (data-object/length 'data->x509-cert data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (let-location ([cert botan-x509-cert #f])
      (case ((foreign-lambda* int ([(c-pointer botan-x509-cert) cert] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count])
               "C_return(botan_x509_cert_load(cert, data + start, count));")
             (location cert) data* from data-length)
        [(0)  (set-finalizer! cert x509-cert-destroy!)]
        [else => (botan-error 'data->x509-cert data from: from to: to)]))))

(define (load-x509-cert path)
  (data->x509-cert
   (call-with-input-file path (cut read-string #f <>))))

(define *x509-cert-begin*
  "-----BEGIN CERTIFICATE-----")
(define *x509-cert-end*
  "-----END CERTIFICATE-----")

(define (load-x509-cert-chain path)
  (let ([data (call-with-input-file path (cut read-string #f <>))])
    (let loop ([start 0] [chain '()])
      (let ([start (string-contains data *x509-cert-begin* start)]
            [stop (string-contains data *x509-cert-end* start)])
        (cond
          [(and start stop)
           (set! stop (+ stop (string-length *x509-cert-end*)))
           (loop stop (cons (data->x509-cert data from: start to: stop) chain))]
          [else
           (reverse! chain)])))))

(define (x509-cert->string cert)
  (let ([s (make-string 4096)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_x509_cert_to_string" botan-x509-cert (scheme-pointer char) (c-pointer size_t))
             cert s (location n))
        [(0)  (substring s 0 (sub1 n))]
        [else => (botan-error 'x509-cert->string cert)]))))

(define (x509-cert->pubkey cert)
  (let-location ([pk botan-pubkey #f])
    (case ((foreign-lambda int "botan_x509_cert_get_public_key" botan-x509-cert (c-pointer botan-pubkey))
           cert (location pk))
      [(0)  (set-finalizer! pk pubkey-destroy!)]
      [else => (botan-error 'x509-cert->pubkey cert)])))

(define (x509-cert-not-before cert)
  (let-location ([time unsigned-integer64 0])
    (case ((foreign-lambda int "botan_x509_cert_not_before" botan-x509-cert (c-pointer unsigned-integer64))
           cert (location time))
      [(0)  time]
      [else => (botan-error 'x509-cert-not-before cert)])))

(define (x509-cert-not-after cert)
  (let-location ([time unsigned-integer64 0])
    (case ((foreign-lambda int "botan_x509_cert_not_after" botan-x509-cert (c-pointer unsigned-integer64))
           cert (location time))
      [(0)  time]
      [else => (botan-error 'x509-cert-not-after cert)])))

(define (x509-cert-fingerprint cert #!key [hash 'SHA-256])
  (when (symbol? hash) (set! hash (symbol->string hash)))
  (let ([s (make-string 1024)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_x509_cert_get_fingerprint" botan-x509-cert nonnull-c-string (scheme-pointer char) (c-pointer size_t))
             cert hash s (location n))
        [(0)  (substring s 0 (sub1 n))]
        [else => (botan-error 'x509-cert-fingerprint cert hash: hash)]))))

(define (x509-cert-serial-number cert)
  (let ([v (make-u8vector 256)])
    (let-location ([n size_t (u8vector-length v)])
      (case ((foreign-lambda int "botan_x509_cert_get_serial_number" botan-x509-cert nonnull-u8vector (c-pointer size_t))
             cert v (location n))
        [(0)  (if (= n (u8vector-length v)) v (subu8vector v 0 n))]
        [else => (botan-error 'x509-cert-serial-number cert)]))))

(define (x509-cert-issuer-key-id cert)
  (let ([v (make-u8vector 256)])
    (let-location ([n size_t (u8vector-length v)])
      (case ((foreign-lambda int "botan_x509_cert_get_authority_key_id" botan-x509-cert nonnull-u8vector (c-pointer size_t))
             cert v (location n))
        [(0)  (if (= n (u8vector-length v)) v (subu8vector v 0 n))]
        [else => (botan-error 'x509-cert-issuer-key-id cert)]))))

(define (x509-cert-issuer-dn cert field #!optional [index 0])
  (let ([s (make-string 256)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_x509_cert_get_issuer_dn" botan-x509-cert symbol size_t (scheme-pointer char) (c-pointer size_t))
             cert field index s (location n))
        [(0)   (substring s 0 (sub1 n))]
        [(-20) #f]
        [else  => (botan-error 'x509-cert-issuer-key-id cert field index)]))))

(define (x509-cert-subject-key-id cert)
  (let ([v (make-u8vector 256)])
    (let-location ([n size_t (u8vector-length v)])
      (case ((foreign-lambda int "botan_x509_cert_get_subject_key_id" botan-x509-cert nonnull-u8vector (c-pointer size_t))
             cert v (location n))
        [(0)  (if (= n (u8vector-length v)) v (subu8vector v 0 n))]
        [else => (botan-error 'x509-cert-subject-key-id cert)]))))

(define (x509-cert-subject-dn cert field #!optional [index 0])
  (let ([s (make-string 256)])
    (let-location ([n size_t (string-length s)])
      (case ((foreign-lambda int "botan_x509_cert_get_subject_dn" botan-x509-cert symbol size_t (scheme-pointer char) (c-pointer size_t))
             cert field index s (location n))
        [(0)   (substring s 0 (sub1 n))]
        [(-20) #f]
        [else  => (botan-error 'x509-cert-subject-key-id cert field index)]))))

(define-pointer-type (botan-x509-crl "botan_x509_crl_struct")
  *botan-x509-crl* x509-crl?)

(define (x509-crl-destroy! crl)
  (case ((foreign-lambda int "botan_x509_crl_destroy" botan-x509-crl) crl)
    [(0)  (void)]
    [else => (botan-error 'x509-crl-destroy! crl)]))

(define (data->x509-crl data #!key [from 0] to)
  (receive (data* data-length) (data-object/length 'data->x509-crl data)
    (set! to (if to (min (max 0 to) data-length) data-length))
    (set! from (min (max 0 from) to))
    (set! data-length (- to from))
    (let-location ([crl botan-x509-crl #f])
      (case ((foreign-lambda* int ([(c-pointer botan-x509-crl) crl] [(scheme-pointer (const unsigned-byte)) data] [size_t start] [size_t count])
               "C_return(botan_x509_crl_load(crl, data + start, count));")
             (location crl) data* from data-length)
        [(0)  (set-finalizer! crl x509-crl-destroy!)]
        [else => (botan-error 'data->x509-crl data from: from to: to)]))))

(define (load-x509-crl path)
  (data->x509-crl
   (call-with-input-file path (cut read-string #f <>))))

(define (x509-usage-allowed? cert . usage)
  (case ((foreign-lambda int "botan_x509_cert_allowed_usage" botan-x509-cert unsigned-int)
         cert
         (fold
          (lambda (flag acc)
            (+ acc
             (case flag
               [(digital-signature) 32768]
               [(non-repudiation)   16384]
               [(key-encipherment)  8192]
               [(data-encipherment) 4096]
               [(key-agreement)     2048]
               [(key-cert-sign)     1024]
               [(crl-sign)          512]
               [(encipher-only)     256]
               [(decipher-only)     128]
               [else                (type-error 'usage flag)])))
          0 usage))
    [(0)  #f]
    [(1)  #t]
    [else => (botan-error 'x509-usage-allowed? cert usage)]))

(define (x509-hostname-match? cert hostname)
  (case ((foreign-lambda int "botan_x509_cert_hostname_match" botan-x509-cert c-string) cert hostname)
    [(0)  #t]
    [(-1) #f]
    [else => (botan-error 'x509-hostname-match? cert hostname)]))

(define (x509-revoked? cert crl)
  (case ((foreign-lambda int "botan_x509_is_revoked" botan-x509-crl botan-x509-cert) crl cert)
    [(0)  #t]
    [(-1) #f]
    [else => (botan-error 'x509-revoked? cert crl)]))

(define %x509-cert-validation-status
  (foreign-lambda c-string "botan_x509_cert_validation_status" int))

(define (->pointer-vector x)
  (cond
    [(not x)
     x]
    [(list? x)
     (let ([v (make-pointer-vector (length x) #f)])
       (let loop ([i 0] [rest x])
         (unless (null? rest)
           (pointer-vector-set! v i (car rest))
           (loop (add1 i) (cdr rest))))
       v)]
    [else
     (make-pointer-vector 1 x)]))

(define (x509-verify cert #!key crl intermediate trusted trusted-path [strength 0] hostname [time (current-seconds)])
  (let-location ([status int 0])
    (case (let ([crl (->pointer-vector crl)]
                [intermediate (->pointer-vector intermediate)]
                [trusted (->pointer-vector trusted)])
            (if crl
              ((foreign-lambda* int ([(c-pointer int) status] [botan-x509-cert cert]
                                     [pointer-vector crl] [size_t ncrl]
                                     [pointer-vector intermediate] [size_t nintermediate]
                                     [pointer-vector trusted] [size_t ntrusted] [c-string trusted_path]
                                     [size_t strength] [c-string hostname] [unsigned-integer64 time])
                 "C_return(botan_x509_cert_verify_with_crl(status, cert, (const botan_x509_cert_t *)intermediate, nintermediate, (const botan_x509_cert_t *)trusted, ntrusted, (const botan_x509_crl_t *)crl, ncrl, trusted_path, strength, hostname, time));")
               (location status) cert
               crl (pointer-vector-length crl)
               intermediate (if intermediate (pointer-vector-length intermediate) 0)
               trusted (if trusted (pointer-vector-length trusted) 0) trusted-path
               strength hostname time)
              ((foreign-lambda* int ([(c-pointer int) status] [botan-x509-cert cert]
                                     [pointer-vector intermediate] [size_t nintermediate]
                                     [pointer-vector trusted] [size_t ntrusted] [c-string trusted_path]
                                     [size_t strength] [c-string hostname] [unsigned-integer64 time])
                 "C_return(botan_x509_cert_verify(status, cert, (const botan_x509_cert_t *)intermediate, nintermediate, (const botan_x509_cert_t *)trusted, ntrusted, trusted_path, strength, hostname, time));")
               (location status) cert
               intermediate (if intermediate (pointer-vector-length intermediate) 0)
               trusted (if trusted (pointer-vector-length trusted) 0) trusted-path
               strength hostname time)))
      [(0)  (values #t (%x509-cert-validation-status status))]
      [(1)  (values #f (%x509-cert-validation-status status))]
      [else => (botan-error/message 'x509-valid? (%x509-cert-validation-status status) cert crl: crl intermediate: intermediate trusted: trusted trusted-path: trusted-path strength: strength hostname: hostname time: time)])))

(define-pointer-type (botan-hotp "botan_hotp_struct")
  *botan-hotp* hotp?)

(define (hotp-destroy! otp)
  (case ((foreign-lambda int "botan_hotp_destroy" botan-hotp) otp)
    [(0)  (void)]
    [else => (botan-error 'hotp-destroy! otp)]))

(define (make-hotp key #!key [hash 'SHA-1] [digits 6])
  (when (symbol? hash) (set! hash (symbol->string hash)))
  (let-location ([otp botan-hotp #f])
    (case ((foreign-lambda int "botan_hotp_init" (c-pointer botan-hotp) nonnull-u8vector size_t nonnull-c-string size_t)
           (location otp) key (u8vector-length key) hash digits)
      [(0)  (set-finalizer! otp hotp-destroy!)]
      [else => (botan-error 'make-hotp hash: hash digits: digits)])))

(define (hotp-generate otp counter)
  (let-location ([code unsigned-integer32 0])
    (case ((foreign-lambda int "botan_hotp_generate" botan-hotp (c-pointer unsigned-integer32) unsigned-integer64)
           otp (location code) counter)
      [(0)  code]
      [else => (botan-error 'hotp-generate otp counter)])))

(define (hotp-valid? otp code counter #!key [resync-range 0])
  (let-location ([next-counter unsigned-integer64 0])
    (case ((foreign-lambda int "botan_hotp_check" botan-hotp (c-pointer unsigned-integer64) unsigned-integer32 unsigned-integer64 size_t)
           otp (location next-counter) code counter resync-range)
      [(0)  next-counter]
      [(1)  #f]
      [else => (botan-error 'hotp-valid? otp code counter resync-range: resync-range)])))

(define-pointer-type (botan-totp "botan_totp_struct")
  *botan-totp* totp?)

(define (totp-destroy! otp)
  (case ((foreign-lambda int "botan_totp_destroy" botan-totp) otp)
    [(0)  (void)]
    [else => (botan-error 'totp-destroy! otp)]))

(define (make-totp key #!key [hash 'SHA-1] [digits 6] [time-step 30])
  (when (symbol? hash) (set! hash (symbol->string hash)))
  (let-location ([otp botan-totp #f])
    (case ((foreign-lambda int "botan_totp_init" (c-pointer botan-totp) nonnull-u8vector size_t nonnull-c-string size_t size_t)
           (location otp) key (u8vector-length key) hash digits time-step)
      [(0)  (set-finalizer! otp totp-destroy!)]
      [else => (botan-error 'make-totp hash: hash digits: digits time-step: time-step)])))

(define (totp-generate otp #!key [time (current-seconds)])
  (let-location ([code unsigned-integer32 0])
    (case ((foreign-lambda int "botan_totp_generate" botan-totp (c-pointer unsigned-integer32) unsigned-integer64)
           otp (location code) time)
      [(0)  code]
      [else => (botan-error 'totp-generate otp time: time)])))

(define (totp-valid? otp code #!key [time (current-seconds)] [acceptable-drift 0])
  (case ((foreign-lambda int "botan_totp_check" botan-totp unsigned-integer32 unsigned-integer64 size_t)
         otp code time acceptable-drift)
    [(0)  #t]
    [(1)  #f]
    [else => (botan-error 'totp-valid? otp code time: time acceptable-drift: acceptable-drift)]))

(define-pointer-type (botan-fpe "botan_fpe_struct")
  *botan-fpe* fpe?)

(define (fpe-destroy! fpe)
  (case ((foreign-lambda int "botan_fpe_destroy" botan-fpe) fpe)
    [(0)  (void)]
    [else => (botan-error 'fpe-destroy! fpe)]))

(define (make-fpe key #!key [limit (expt 2 30)] [rounds 5] fe1-compat)
  (let-location ([fpe botan-fpe #f])
    (case ((foreign-lambda int "botan_fpe_fe1_init" (c-pointer botan-fpe) botan-mp nonnull-u8vector size_t size_t unsigned-int32)
	   (location fpe) (%number->mp 'make-fpe limit) key (u8vector-length key) rounds (if fe1-compat 1 0))
      [(0)  (set-finalizer! fpe fpe-destroy!)]
      [else => (botan-error 'make-fpe limit: limit rounds: rounds fe1-compat: fe1-compat)])))

(define (fpe-encrypt fpe x tweak)
  (let ([x (%number->mp 'fpe-encrypt x)])
    (case ((foreign-lambda int "botan_fpe_encrypt" botan-fpe botan-mp nonnull-u8vector size_t)
	   fpe x tweak (u8vector-length tweak))
      [(0)  (%mp->number 'fpe-encrypt x)]
      [else => (botan-error 'fpe-encrypt fpe x tweak)])))

(define (fpe-decrypt fpe x tweak)
  (let ([x (%number->mp 'fpe-decrypt x)])
    (case ((foreign-lambda int "botan_fpe_decrypt" botan-fpe botan-mp nonnull-u8vector size_t)
	   fpe x tweak (u8vector-length tweak))
      [(0)  (%mp->number 'fpe-decrypt x)]
      [else => (botan-error 'fpe-decrypt fpe x tweak)])))

)

;; vim: set ai et ts=4 sts=2 sw=2: ;;
